/*
    Kernel mode HTTP server for Windows.
    Copyright (C) 2009-2015 Bo Brantn.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <ntddk.h>
#include "fileio.h"
#include "ksocket.h"

#define KHTTPD_POOL_TAG             'ptth'
#define PARAMETER_KEY               L"\\Parameters"
#define PORT_NUMBER_VALUE           L"Port"
#define CLIENT_PORT_NUMBER_VALUE    L"ClientPort"
#define DOCUMENT_ROOT_VALUE         L"DocumentRoot"
#define DEFAULT_PORT_NUMBER         80
#define DEFAULT_CLIENT_PORT_NUMBER  8080
#define DEFAULT_DOCUMENT_ROOT       L"\\??\\c:\\www"

#define HEADER_HTML     "HTTP/1.0 200 OK\r\nContent-type: text/html\r\n\r\n"
#define HEADER_IMAGE    "HTTP/1.0 200 OK\r\nContent-type: image/jpeg\r\n\r\n"
#define HEADER_BIN      "HTTP/1.0 200 OK\r\nContent-type: application/bin\r\n\r\n"
#define RESPONSE_404    "HTTP/1.0 404 Not Found\r\n\r\nError: 404 Not Found"

ULONG           portNumber;
ULONG           clientPortNumber;
UNICODE_STRING  documentRootName;
HANDLE          documentRoot;

#pragma code_seg("PAGE")

int create_thread(PKSTART_ROUTINE thread_func, void *thread_arg)
{
    NTSTATUS    status;
    HANDLE      thread_handle;

    PAGED_CODE();

    if (thread_func == NULL) {
        return -1;
    }

    status = PsCreateSystemThread(
        &thread_handle,
        (ACCESS_MASK) 0L,
        NULL,
        NULL,
        NULL,
        (PKSTART_ROUTINE) thread_func,
        thread_arg
        );

    if (!NT_SUCCESS(status)) {
        return status;
    }

    ZwClose(thread_handle);

    return STATUS_SUCCESS;
}

static void slash_to_backslash(char *str)
{
    char *p;

    PAGED_CODE();

    for (p = str; *p != '\0'; p++)
    {
        if (*p == '/')
        {
            *p = '\\';
        }
    }
}

KSTART_ROUTINE request_thread;

void request_thread(void *arg)
{
    INT_PTR reqSocket = (INT_PTR) arg;
    char buf[512];
    int len;
    char *fileName, *end, *ext, *header;
    HANDLE fileHandle;

    PAGED_CODE();

    KdPrint(("khttpd: request_thread\n"));

    len = recv(reqSocket, buf, sizeof(buf), 0);

    if (len < 6  && strncmp(buf, "GET /", 5) != 0)
    {
        return;
    }

    fileName = &buf[5];
    end = strchr(fileName, ' ');
    if (!end)
    {
        return;
    }
    *end = '\0';

    slash_to_backslash(fileName);

    fileHandle = fopen(fileName, documentRoot);

    if (!fileHandle)
    {
        KdPrint(("khttpd: 404 Not Found %s\n", fileName));
        send_all(reqSocket, RESPONSE_404, sizeof(RESPONSE_404) - 1, 0);
        return;
    }

    KdPrint(("khttpd: 200 OK %s\n", fileName));

    ext = fileName + strlen(fileName) - 4;

    if (strcmp(ext, ".htm") == 0)
    {
        header = HEADER_HTML;
    }
    else if (strcmp(ext, ".jpg") == 0)
    {
        header = HEADER_IMAGE;
    }
    else
    {
        header = HEADER_BIN;
    }

    //send_file(reqSocket, header, fileHandle);
    //send_file_mdl(reqSocket, header, fileHandle);
    send_file_mn_mdl(reqSocket, header, fileHandle);
    //send_file_mn_mdl_async(reqSocket, header, fileHandle);

    close(reqSocket);
    fclose(fileHandle);

    KdPrint(("khttpd: request_thread done\n"));

    return;
}

KSTART_ROUTINE main_thread;

void main_thread(void *arg)
{
    INT_PTR             srvSocket = (INT_PTR) arg, reqSocket;
    struct sockaddr_in  remoteAddr;
    int                 remoteLen = sizeof(remoteAddr);
#if DBG
    char                addrStr[16];
#endif

    PAGED_CODE();

    while ((reqSocket = accept(srvSocket, (struct sockaddr*) &remoteAddr, &remoteLen)) != -1)
    {
#if DBG
        inet_ntoa_r(remoteAddr.sin_addr, addrStr);
        DbgPrint("khttpd: connection from %s:%u\n", addrStr, ntohs(remoteAddr.sin_port));
#endif
        create_thread(request_thread, (void *) reqSocket);
    }

    DbgPrint("khttpd: accept() error, terminating main thread\n");
    return;
}

#pragma code_seg("INIT")

static NTSTATUS queryRegistry(PUNICODE_STRING registryPath)
{
    UNICODE_STRING              parameterPath;
    RTL_QUERY_REGISTRY_TABLE    queryTable[2];
    NTSTATUS                    status;

    parameterPath.Length = 0;

    parameterPath.MaximumLength = registryPath->Length + sizeof(PARAMETER_KEY);

    parameterPath.Buffer = (PWSTR) ExAllocatePoolWithTag(PagedPool, parameterPath.MaximumLength, KHTTPD_POOL_TAG);

    if (parameterPath.Buffer == NULL)
    {
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    RtlCopyUnicodeString(&parameterPath, registryPath);

    RtlAppendUnicodeToString(&parameterPath, PARAMETER_KEY);

    RtlZeroMemory(&queryTable[0], sizeof(queryTable));

    queryTable[0].Flags = RTL_QUERY_REGISTRY_DIRECT | RTL_QUERY_REGISTRY_REQUIRED;
    queryTable[0].Name = PORT_NUMBER_VALUE;
    queryTable[0].EntryContext = &portNumber;

    status = RtlQueryRegistryValues(
        RTL_REGISTRY_ABSOLUTE,
        parameterPath.Buffer,
        &queryTable[0],
        NULL,
        NULL
        );

    if (!NT_SUCCESS(status))
    {
        DbgPrint("khttpd: Query registry failed, using default port number.\n");
        portNumber = DEFAULT_PORT_NUMBER;
    }

    RtlZeroMemory(&queryTable[0], sizeof(queryTable));

    queryTable[0].Flags = RTL_QUERY_REGISTRY_DIRECT | RTL_QUERY_REGISTRY_REQUIRED;
    queryTable[0].Name = CLIENT_PORT_NUMBER_VALUE;
    queryTable[0].EntryContext = &clientPortNumber;

    status = RtlQueryRegistryValues(
        RTL_REGISTRY_ABSOLUTE,
        parameterPath.Buffer,
        &queryTable[0],
        NULL,
        NULL
        );

    if (!NT_SUCCESS(status))
    {
        DbgPrint("khttpd: Query registry failed, using default client port number.\n");
        clientPortNumber = DEFAULT_CLIENT_PORT_NUMBER;
    }

    RtlZeroMemory(&queryTable[0], sizeof(queryTable));

    queryTable[0].Flags = RTL_QUERY_REGISTRY_DIRECT | RTL_QUERY_REGISTRY_REQUIRED;
    queryTable[0].Name = DOCUMENT_ROOT_VALUE;
    queryTable[0].EntryContext = &documentRootName;

    RtlZeroMemory(&documentRootName, sizeof(documentRootName));

    status = RtlQueryRegistryValues(
        RTL_REGISTRY_ABSOLUTE,
        parameterPath.Buffer,
        &queryTable[0],
        NULL,
        NULL
        );

    if (!NT_SUCCESS(status))
    {
        DbgPrint("khttpd: Query registry failed, using default document root.\n");
        RtlInitUnicodeString(&documentRootName, DEFAULT_DOCUMENT_ROOT);
    }

    ExFreePool(parameterPath.Buffer);

    return STATUS_SUCCESS;
}

DRIVER_INITIALIZE DriverEntry;

NTSTATUS
DriverEntry(
    IN PDRIVER_OBJECT   DriverObject,
    IN PUNICODE_STRING  RegistryPath
    )
{
    NTSTATUS            status;
    UNICODE_STRING      deviceName;
    PDEVICE_OBJECT      deviceObj;
    INT_PTR             srvSocket;
    struct sockaddr_in  localAddr;
    int                 rVal;

    KdPrint(("khttpd: DriverEntry\n"));

    status = queryRegistry(RegistryPath);

    if (!NT_SUCCESS(status))
    {
        DbgPrint("khttpd: Query registry error %#x\n", status);
        return status;
    }

    KdPrint(("khttpd: port number %u\n", portNumber));
    KdPrint(("khttpd: client port number %u\n", clientPortNumber));
    KdPrint(("khttpd: document root %.*S\n", documentRootName.Length, documentRootName.Buffer));

    RtlInitUnicodeString(&deviceName, L"\\Device\\Khttpd");

    status = IoCreateDevice(
        DriverObject,
        0,
        &deviceName,
        FILE_DEVICE_UNKNOWN,
        0,
        FALSE,
        &deviceObj
        );

    if (!NT_SUCCESS(status))
    {
        DbgPrint("khttpd: IoCreateDevice error %#x\n", status);
        return status;
    }

    documentRoot = dopen(&documentRootName);

    if (!documentRoot)
    {
        DbgPrint("khttpd: error open document root %.*S\n", documentRootName.Length, documentRootName.Buffer);
        IoDeleteDevice(deviceObj);
        return STATUS_UNSUCCESSFUL;
    }

    srvSocket = socket(AF_INET, SOCK_STREAM, 0);

    if (srvSocket == -1)
    {
        DbgPrint("khttpd: socket() error\n");
        fclose(documentRoot);
        IoDeleteDevice(deviceObj);
        return STATUS_UNSUCCESSFUL;
    }

    localAddr.sin_family = AF_INET;
    localAddr.sin_port = htons((u_short) portNumber);
    localAddr.sin_addr.s_addr = INADDR_ANY;

    rVal = bind(srvSocket, (struct sockaddr*) &localAddr, sizeof(localAddr));

    if (rVal < 0)
    {
        DbgPrint("khttpd: bind error %#x\n", rVal);
        close(srvSocket);
        fclose(documentRoot);
        IoDeleteDevice(deviceObj);
        return STATUS_UNSUCCESSFUL;
    }

    rVal = listen(srvSocket, SOMAXCONN);

    if (rVal < 0)
    {
        DbgPrint("khttpd: listen error %#x\n", rVal);
        close(srvSocket);
        fclose(documentRoot);
        IoDeleteDevice(deviceObj);
        return STATUS_UNSUCCESSFUL;
    }

    create_thread(main_thread, (void*) srvSocket);

    KdPrint(("khttpd: DriverEntry done\n"));

    return STATUS_SUCCESS;
}

#pragma code_seg() // end "INIT"
